class Redshift:
    def __init__(self, redshift=0.0):
        self.redshift = redshift


class SphProfile:
    def __init__(self, centre=(0.0, 0.0)):
        """Generic circular profiles class to contain functions shared by light and
        mass profiles.

        Parameters
        ----------
        centre
            The (y,x) coordinates of the origin of the profile.
        """
        self.centre = centre


class EllProfile(SphProfile):
    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """Generic elliptical profiles class to contain functions shared by light
        and mass profiles.

        Parameters
        ----------
        centre
            The (y,x) coordinates of the origin of the profiles
        axis_ratio
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllProfile, self).__init__(centre)
        self.axis_ratio = axis_ratio
        self.phi = phi


class Gaussian(EllProfile):
    def __init__(
        self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, sigma=0.01
    ):
        """The elliptical Gaussian profile.

        Parameters
        ----------
        centre
            The (y,x) origin of the light profile.
        axis_ratio
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity
            Overall intensity normalisation of the light profiles (electrons per
            second).
        sigma
            The full-width half-maximum of the Gaussian.
        """
        super(Gaussian, self).__init__(centre, axis_ratio, phi)

        self.intensity = intensity
        self.sigma = sigma



import inspect
import json
import logging

import numpy as np

from autoconf.class_path import get_class_path, get_class

logger = logging.getLogger(__name__)


def nd_array_as_dict(obj: np.ndarray) -> dict:
    """
    Converts a numpy array to a dictionary representation.
    """
    return {
        "type": "numpy.ndarray",
        "array": obj.tolist(),
        "dtype": str(obj.dtype),
    }


def nd_array_from_dict(nd_array_dict: dict) -> np.ndarray:
    """
    Converts a dictionary representation back to a numpy array.
    """
    return np.array(nd_array_dict["array"], dtype=getattr(np, nd_array_dict["dtype"]))


def as_dict(obj):
    if isinstance(obj, np.ndarray):
        try:
            return nd_array_as_dict(obj)
        except Exception as e:
            logger.info(e)

    if isinstance(obj, list):
        return list(map(as_dict, obj))
    if isinstance(obj, dict):
        return {
            "type": "dict",
            **{key: as_dict(value) for key, value in obj.items()},
        }
    if obj.__class__.__module__ == "builtins":
        return obj
    argument_dict = {
        arg: getattr(obj, arg) for arg in inspect.getfullargspec(obj.__init__).args[1:]
    }

    return {
        "type": get_class_path(obj.__class__),
        **{key: as_dict(value) for key, value in argument_dict.items()},
    }


class Dictable:
    def dict(self) -> dict:
        """
        A dictionary representation of the instance comprising a type
        field which contains the entire class path by which the type
        can be imported and constructor arguments.
        """
        # noinspection PyTypeChecker
        return as_dict(self)

    @staticmethod
    def from_dict(cls_dict):
        """
        Instantiate an instance of a class from its dictionary representation.

        Parameters
        ----------
        cls_dict
            A dictionary representation of the instance comprising a type
            field which contains the entire class path by which the type
            can be imported and constructor arguments.

        Returns
        -------
        An instance of the geometry profile specified by the type field in
        the cls_dict
        """
        if isinstance(cls_dict, list):
            return list(map(Dictable.from_dict, cls_dict))
        if not isinstance(cls_dict, dict):
            return cls_dict

        cls = get_class(cls_dict.pop("type"))

        if cls is np.ndarray:
            return nd_array_from_dict(cls_dict)

        # noinspection PyArgumentList
        return cls(
            **{name: Dictable.from_dict(value) for name, value in cls_dict.items()}
        )

    @classmethod
    def from_json(cls, file_path: str) -> "Dictable":
        """
        Load the dictable object to a .json file, whereby all attributes are converted from the .json file's dictionary
        representation to create the instance of the object

        A json file of the instance can be created from the .json file via the `output_to_json` method.

        Parameters
        ----------
        file_path
            The path to the .json file that the dictionary representation of the object is loaded from.
        """
        with open(file_path, "r+") as f:
            cls_dict = json.load(f)

        return cls.from_dict(cls_dict)

    def output_to_json(self, file_path: str):
        """
        Output the dictable object to a .json file, whereby all attributes are converted to a dictionary representation
        first.

        An instane of the object can be created from the .json file via the `from_json` method.

        Parameters
        ----------
        file_path
            The path to the .json file that the dictionary representation of the object is written too.
        """
        with open(file_path, "w+") as f:
            json.dump(self.dict(), f, indent=4)

    def __eq__(self, other):
        return self.dict() == other.dict()



class CachedProperty(object):
    """
    A property that is only computed once per instance and then replaces
    itself with an ordinary attribute. Deleting the attribute resets the
    property.

    Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


cached_property = CachedProperty



import inspect
import json
import logging
import os
from os import path
from importlib import util
from pathlib import Path

from .config import make_config_for_class

logger = logging.getLogger(__name__)


def for_file(module_path: str) -> dict:
    """
    Generate JSON priors for all classes in a file, using default
    prior configuration for each constructor argument.

    Parameters
    ----------
    module_path
        The path to the file.

    Returns
    -------
    JSON configuration, where class names are mapped to their prior configs.
    """
    spec = util.spec_from_file_location("module.name", module_path)
    module = util.module_from_spec(spec)
    spec.loader.exec_module(module)
    classes = inspect.getmembers(module)

    return {
        name: make_config_for_class(obj)[1]
        for name, obj in classes
        if inspect.isclass(obj)
    }


def generate(directory: str):
    """
    Generate prior configuration for a given directory, recursively.

    A directory "priors" is created if it does not exists. A new JSON file is created
    in priors for each python module found that contains at least one class.

    If an output file already exists then prior generation is skipped.

    Parameters
    ----------
    directory
        The directory for which prior are generated
    """
    cwd = Path(os.getcwd())
    try:
        os.mkdir(path.join(cwd, "priors"))
    except FileExistsError:
        pass
    for directory, _, files in os.walk(directory):
        directory = Path(directory)
        for file in files:
            if file.endswith(".py"):
                full_path = directory / file
                spec = for_file(full_path)
                config_path = cwd / "priors" / file.replace(".py", ".json")
                if len(spec) > 0:
                    if os.path.exists(config_path):
                        logger.info(f"{config_path} already exists")
                        continue
                    with open(config_path, "w+") as f:
                        json.dump(spec, f)



import inspect
import json
import logging
from collections.abc import Sized
from glob import glob
from pathlib import Path
from typing import List, Type, Tuple
import ntpath
from os import path

import yaml

from autoconf.directory_config import family

logger = logging.getLogger(__name__)

default_prior = {
    "type": "Uniform",
    "lower_limit": 0.0,
    "upper_limit": 1.0,
    "width_modifier": {"type": "Absolute", "value": 0.2},
    "gaussian_limits": {"lower": 0.0, "upper": 1.0},
}


def make_config_for_class(cls):
    path = path_for_class(cls)
    arg_spec = inspect.getfullargspec(cls)
    arguments = arg_spec.args[1:]
    defaults = list(reversed(arg_spec.defaults or list()))

    config = dict()
    for i, argument in enumerate(reversed(arguments)):
        if i < len(defaults):
            default = defaults[i]
            if isinstance(default, Sized):
                for j in range(len(default)):
                    config[f"{argument}_{j}"] = default_prior
                continue
        config[argument] = default_prior

    return path, config


def path_for_class(cls) -> List[str]:
    """
    A list describing the import path for a given class.

    Parameters
    ----------
    cls
        A class with some module path

    Returns
    -------
    A list of modules terminating in the name of a class
    """
    return f"{cls.__module__}.{cls.__name__}".split(".")


class JSONPriorConfig:
    def __init__(self, config_dict: dict, directory=None):
        """
        Parses configuration describing priors associated with classes.

        The path pointing to a class is the same as the path to import it.

        Paths can be strings with '.' as a delimiter.
        {"module.class": config}

        Else they can be a series of dictionary keys.
        {"module": {"class": config}}

        Or any combination thereof.

        Parameters
        ----------
        config_dict
            A dictionary describing the prior configuration for constructor arguments
            of different classes.
        """
        self.obj = config_dict
        self.directory = directory
        self._path_value_map = None

    @property
    def paths(self):
        return list(self.path_value_map.keys())

    @property
    def path_value_map(self) -> dict:
        """
        A dictionary matching every possible path to the configuration it points to.
        """
        if self._path_value_map is None:

            def get_path_values(obj):
                path_values = dict()
                if isinstance(obj, dict):
                    for key, value in obj.items():
                        path_values[key] = value
                        for path, path_value in get_path_values(value).items():
                            path_values[f"{key}.{path}"] = path_value

                return path_values

            self._path_value_map = get_path_values(self.obj)
        return self._path_value_map

    @property
    def path_value_tuples(self) -> List[Tuple[str, object]]:
        """
        Tuple pairs matching every possible path to the configuration it points to.
        These are ordered by key length with the longest key first.
        """
        return sorted(
            list(self.path_value_map.items()),
            key=lambda item: len(item[0]),
            reverse=True,
        )

    @classmethod
    def from_directory(cls, directory: str) -> "JSONPriorConfig":
        """
        Load JSONPriorConfiguration from a file.

        Parameters
        ----------
        directory
            The path to a file.

        Returns
        -------
        A configuration instance.
        """
        config_dict = dict()

        config_path = Path(directory)

        for suffix, parser in [
            ("json", json.load),
            ("yaml", yaml.safe_load),
            ("yml", yaml.safe_load),
        ]:
            for file in config_path.rglob(f"*.{suffix}"):
                parts = file.relative_to(config_path).with_suffix("").parts
                with open(file) as f:
                    config_dict[".".join(parts)] = parser(f)

        return JSONPriorConfig(config_dict, directory=directory)

    def __str__(self):
        return json.dumps(self.obj)

    def __getitem__(self, item):
        return JSONPriorConfig(self.obj[".".join(item)], directory=self.directory)

    def __contains__(self, item):
        return ".".join(item) in self.obj

    def for_class_and_suffix_path(self, cls: Type, suffix_path: List[str]):
        """
        Get configuration for a prior.

        If it is just basic configuration then the suffix path is just the
        name of the prior in a list. Width configuration also adds an
        additional "width_modifier" item.

        If configuration is not found for the class then configurations for
        parents of the class are searched.

        Parameters
        ----------
        cls
            The class with which the prior is associated.
        suffix_path
            The path to the prior.

        Returns
        -------
        A configuration dictionary
        """
        for c in family(cls):
            try:
                return self(path_for_class(c) + suffix_path)
            except KeyError:
                pass
        raise KeyError(
            f"No config found for class {cls} and path {suffix_path} in {self.directory}"
        )

    def __call__(self, config_path: List[str]):
        """
        Get the config at the end of the config_path.

        The configuration dictionary is traversed until config is found
        at the end, else an exception is thrown.

        Parameters
        ----------
        config_path
            The import path of a package, module, class or class and constructor
            argument name.

        Returns
        -------
        A configuration dictionary or value

        Raises
        ------
        PriorException
            If no configuration is found.
        """
        key = ".".join(config_path)
        for path, value in self.path_value_tuples:
            if key.endswith(path):
                return value
        raise KeyError(
            f"No configuration was found for the path {config_path}"
            + ("" if self.directory is None else f" ({self.directory})")
        )



import logging
import logging.config
import os
import shutil
from functools import wraps
from pathlib import Path
from typing import Optional, Union, Dict, MutableMapping

import yaml

from autoconf.directory_config import (
    RecursiveConfig,
    PriorConfigWrapper,
    AbstractConfig,
    family,
)
from autoconf.exc import ConfigException
from autoconf.json_prior.config import JSONPriorConfig

logger = logging.getLogger(__name__)

LOGGING_CONFIG_FILE = "logging.yaml"


def get_matplotlib_backend():
    try:
        return instance["visualize"]["general"]["general"]["backend"]
    except KeyError:
        return "default"


class DictWrapper(MutableMapping):
    def __delitem__(self, v) -> None:
        del self._dict[v]

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def __init__(self, paths):
        self._dict = dict()
        self.paths = paths

    def __contains__(self, item):
        return item.lower() in self._dict

    def items(self):
        return self._dict.items()

    def __setitem__(self, key, value):
        if isinstance(key, str):
            key = key.lower()
        self._dict[key] = value

    def __getitem__(self, key):
        if isinstance(key, str):
            key = key.lower()
        try:
            return self._dict[key]
        except KeyError:
            raise KeyError(f"key {key} not found in paths {self.paths_string}")

    @property
    def paths_string(self):
        return "\n".join(map(str, self.paths))

    def __repr__(self):
        return repr(self._dict)

    def family(self, cls):
        for item in family(cls):
            try:
                return self[item]
            except KeyError:
                pass
        raise KeyError(
            f"config for {cls} or its parents not found in paths {self.paths_string}"
        )


class Config:
    def __init__(self, *config_paths, output_path: Union[str, Path] = "output"):
        """
        Singleton to manage configuration.

        Configuration is loaded using the __getitem__ syntax where the key entered
        can refer to a directory, file, section or item.

        Configuration is first attempted to be loaded from the directory indicated by the first
        config_path. If no configuration is found the second directory is searched and so on.
        This allows a default configuration to be defined with additional configuration overriding
        it.

        Parameters
        ----------
        config_paths
            Indicate directories where configuration is defined, in the order of priority with
            configuration in the first config_path overriding configuration in later config
            paths
        output_path
            The path where data should be saved.
        """
        for config_path in config_paths:
            if Path(config_path).name == "output":
                logger.warning(
                    f"{config_path} passed as config path. Did you mean to use output_path={config_path}?"
                )

        self._prior_config = None

        self._configs = list()
        self._dict = DictWrapper(self.paths)

        self.configs = list(map(RecursiveConfig, config_paths))

        self.output_path = output_path

    @property
    def dict(self):
        """
        A dictionary containing configuration loaded from directories and files.

        Lazily recurse directories to load configuration.
        """
        if self._dict is None:
            self._dict = DictWrapper(self.paths)

            def recurse_config(config, d):
                try:
                    for key, value in config.items():
                        if isinstance(value, AbstractConfig):
                            if key not in d:
                                d[key] = DictWrapper(self.paths)
                            recurse_config(value, d=d[key])
                        else:
                            d[key] = value
                except KeyError as e:
                    logger.debug(e)

            for config_ in reversed(self._configs):
                recurse_config(config_, self._dict)
        return self._dict

    def configure_logging(self):
        """
        Set the most up to date logging configuration
        """
        logging_config = self.logging_config
        if logging_config is not None:
            logging.config.dictConfig(logging_config)

    @property
    def logging_config(self) -> Optional[Dict]:
        """
        Loading logging configuration from a YAML file
        from the most recently added config directory
        for which it exists.
        """
        for config in self.configs:
            path = config.path
            try:
                if LOGGING_CONFIG_FILE in os.listdir(config.path):
                    with open(path / LOGGING_CONFIG_FILE) as f:
                        return yaml.safe_load(f)
            except FileNotFoundError:
                logger.debug(f"No configuration found at path {config.path}")
        return None

    @property
    def configs(self):
        return self._configs

    @configs.setter
    def configs(self, configs):
        """
        When the list of configs is updated this invalidates the current config
        dictionary
        """
        self._prior_config = None
        self._dict = None
        self._configs = configs

    def __getitem__(self, item):
        return self.dict[item]

    def __iter__(self):
        return iter(self.dict)

    @property
    def paths(self):
        return [config.path for config in self._configs]

    @property
    def prior_config(self) -> PriorConfigWrapper:
        """
        Configuration for priors. This indicates, for example, the mean and the width of priors
        for the attributes of given classes.
        """
        if self._prior_config is None:
            self._prior_config = PriorConfigWrapper(
                [JSONPriorConfig.from_directory(path / "priors") for path in self.paths]
            )
        return self._prior_config

    def push(
        self,
        new_path: Union[str, Path],
        output_path: Optional[str] = None,
        keep_first: bool = False,
    ):
        """
        Push a new configuration path. This overrides the existing config
        paths, with existing configs being used as a backup when a value
        cannot be found in an overriding config.

        Parameters
        ----------
        new_path
            A path to config directory
        output_path
            The path at which data should be output. If this is None then it remains
            unchanged
        keep_first
            If True the current priority configuration mains such.

        Raises
        ------
        ConfigException
            If the pushed path does not exist or does not contain at least one file
            with an expected configuration suffix
        """
        logger.debug(f"Pushing new config with path {new_path}")

        if not Path(new_path).exists():
            raise ConfigException(f"{new_path} does not exist")

        suffixes = {
            Path(file).suffix for _, _, files in os.walk(new_path) for file in files
        }

        CONFIG_SUFFIXES = [
            ".yml",
            ".ini",
            ".json",
            ".yaml",
        ]

        if not any((suffix in suffixes for suffix in CONFIG_SUFFIXES)):
            raise ConfigException(
                f"{new_path} does not contain any files ending with {'/'.join(CONFIG_SUFFIXES)} recursively"
            )

        self.output_path = output_path or self.output_path

        try:
            if self.configs[0] == new_path or (
                keep_first and len(self.configs) > 1 and self.configs[1] == new_path
            ):
                return
        except IndexError:
            pass

        new_config = RecursiveConfig(new_path)

        configs = list(filter(lambda config: config != new_config, self.configs))
        if keep_first:
            self.configs = configs[:1] + [new_config] + configs[1:]
        else:
            self.configs = [new_config] + configs

        self.configure_logging()

    def register(self, file: str):
        """
        Add defaults for a given project

        Parameters
        ----------
        file
            The path to the project's __init__
        """
        self.push(Path(file).parent / "config", keep_first=True)


current_directory = Path(os.getcwd())

default = Config(
    current_directory / "config", output_path=current_directory / "output/"
)

instance = default


def output_path_for_test(temporary_path="temp", remove=True):
    """
    Temporarily change the output path for the scope of a function
    (e.g. a test). Remove the files after the test has completed
    execution.

    Parameters
    ----------
    temporary_path
        The path to temporarily output files to
    remove
        Should the path be removed?

    Returns
    -------
    The original function, decorated
    """

    def remove_():
        if remove:
            shutil.rmtree(temporary_path, ignore_errors=True)

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            remove_()
            original_path = instance.output_path
            instance.output_path = temporary_path

            result = func(*args, **kwargs)

            remove_()
            instance.output_path = original_path

            return result

        return wrapper

    return decorator


def with_config(*path: str, value):
    """
    Create a decorator that swaps a value in configuration
    defined by path for the scope of a test.

    Parameters
    ----------
    path
        A path through config. e.g. "general", "output", "identifier_version"
    value
        The value to temporarily set for the config field

    Returns
    -------
    A decorator
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            config = instance
            for string in path[:-1]:
                config = config[string]

            original_value = config[path[-1]]
            config[path[-1]] = value

            result = func(*args, **kwargs)

            config[path[-1]] = original_value

            return result

        return wrapper

    return decorator



from . import exc
from .tools.decorators import cached_property
from .conf import Config
from .conf import instance
from .json_prior.config import default_prior
from .json_prior.config import make_config_for_class
from .json_prior.config import path_for_class
from .json_prior.config import JSONPriorConfig

__version__ = "2023.6.12.5"
